#dialect "staticTypes"
import "exp" as exp
type Integer = exp.Integer
type Exp = exp.Exp
type Value = exp.Value

// define the Object Algebra machinery

type IntAlg<A> = {
    lit(x:Integer) -> A
    add(e1:A, e2:A) -> A
}

factory method intFactory {
    method lit(x:Integer) -> Exp  { exp.lit(x) }
    method add(e1:Exp, e2:Exp) -> Exp { exp.add(e1, e2) }
}

method make3Plus5<A>(f:IntAlg<A>) -> A {
    f.add(f.lit(3), f.lit(5))
}
// compare with the normal expression:
// def threePlusFour:Exp = add(lit 3, lit 4)

def threePlusFive = make3Plus5 (intFactory)
print(threePlusFive)

// add printing to expressions "retroactively"

type IPrint = {
    print -> String
}

factory method intPrint -> IntAlg<IPrint> {
    factory method lit(x:Integer) {
        method print -> String { x.asString }
    }
    factory method add(e1:IPrint, e2:IPrint) {
        method print -> String { "{e1.print} + {e2.print}" }
    }
}

method threePlusFourExp<A>(v:IntAlg<A>) -> A {
    v.add(v.lit(3), v.lit(4))
}

def x = threePlusFourExp(intFactory).eval
def s = threePlusFourExp(intPrint).print
print "{s} = {x.asInt}"


